import asyncdispatch, bcrypt, cliutils, jester, json, jwt, logging,
  os, osproc, sequtils, strutils, tempfile, times, unittest, uuids

import ./configuration, ./core

type
  Session = object
    user*: UserRef
    issuedAt*, expires*: Time

#const ISO_TIME_FORMAT = "yyyy-MM-dd'T'HH:mm:ss"
const JSON = "application/json"

proc newSession*(user: UserRef): Session =
  result = Session(
    user: user,
    issuedAt: getTime(),
    expires: daysForward(7).toTime())

proc makeJsonResp(status: HttpCode, details: string = ""): string =
  result = $(%* {
    "statusCode": status.int,
    "status": $status,
    "details": details
  })

proc toJWT*(cfg: StrawBossConfig, session: Session): string =
  ## Make a JST token for this session.
  var jwt = JWT(
    header: JOSEHeader(alg: HS256, typ: "jwt"),
    claims: toClaims(%*{
      "sub": session.user.name,
      "iat": session.issuedAt.toSeconds().int,
      "exp": session.expires.toSeconds().int }))

  jwt.sign(cfg.authSecret)
  result = $jwt

proc fromJWT*(cfg: StrawBossConfig, strTok: string): Session =
  ## Validate a given JWT and extract the session data.
  let jwt = toJWT(strTok)
  var secret = cfg.authSecret
  if not jwt.verify(secret): raiseEx "Unable to verify auth token."
  jwt.verifyTimeClaims()

  # Find the user record (if authenticated)
  let username = jwt.claims["sub"].node.str
  let users = cfg.users.filterIt(it.name == username)
  debug "username: " & username & "\n\tusers: " & $users.mapIt(it.name) & "\n\tall users: " & cfg.users.mapIt(it.name)
  if users.len != 1: raiseEx "Could not find session user."

  result = Session(
    user: users[0],
    issuedAt: fromSeconds(jwt.claims["iat"].node.num),
    expires: fromSeconds(jwt.claims["exp"].node.num))

proc extractSession(cfg: StrawBossConfig, request: Request): Session =
  ## Helper to extract a session from a reqest.

  # Find the auth header
  if not request.headers.hasKey("Authorization"):
    raiseEx "No auth token."

  # Read and verify the JWT token
  let headerVal = request.headers["Authorization"]
  if not headerVal.startsWith("Bearer "):
    raiseEx "Invalid Authentication type (only 'Bearer' is supported)."

  result = fromJWT(cfg, headerVal[7..^1])

proc hashPwd*(pwd: string, cost: int8): string =
  let salt = genSalt(cost)
  result = hash(pwd, salt)

proc validatePwd*(u: UserRef, givenPwd: string): bool =
  let salt = u.hashedPwd[0..28] # TODO: magic numbers
  result = compare(u.hashedPwd, hash(givenPwd, salt))

proc makeAuthToken*(cfg: StrawBossConfig, uname, pwd: string): string =
  ## Given a username and pwd, validate the combination and generate a JWT
  ## token string.

  if uname == nil or pwd == nil:
    raiseEx "fields 'username' and 'password' required"

  # find the user record
  let users = cfg.users.filterIt(it.name == uname)
  if users.len != 1: raiseEx "invalid username or password"

  let user = users[0]

  if not validatePwd(user, pwd): raiseEx "invalid username or password"

  let session = newSession(user)

  result = toJWT(cfg, session)

proc makeApiKey*(cfg: StrawBossConfig, uname: string): string =
  ## Given a username, make an API token (JWT token string that does not
  ## expire). Note that this does not validate the username/pwd combination. It
  ## is not intended to be exposed publicly via the API, but serve as a utility
  ## function for an administrator to setup a unsupervised account (git access
  ## for example).

  if uname == nil: raiseEx "no username given"

  # find the user record
  let users = cfg.users.filterIt(it.name == uname)
  if users.len != 1: raiseEx "invalid username"

  let session = Session(
    user: users[0],
    issuedAt: getTime(),
    expires: daysForward(365 * 1000).toTime())

  result = toJWT(cfg, session);

template checkAuth() =
  ## Check this request for authentication and authorization information.
  ## Injects two variables into the running context: the session and authed:
  ## true if the request is authorized, false otherwise. If the request is not
  ## authorized, this template sets up the 401 response correctly. The calling
  ## context needs only to return from the route.

  var session {.inject.}: Session
  var authed {.inject.} = false

  try:
    session = extractSession(cfg, request)
    authed = true
  except:
    debug "Auth failed: " & getCurrentExceptionMsg()
    response.data[2]["WWW-Authenticate"] = "Bearer"
    resp(Http401, makeJsonResp(Http401), JSON)

proc start*(cfg: StrawBossConfig): void =

  var stopFuture = newFuture[void]()
  var workers: seq[Worker] = @[]

  settings:
    port = Port(8180)
    appName = "/api"

  routes:

    post "/auth-token":
      var uname, pwd: string
      try:
        let jsonBody = parseJson(request.body)
        uname = jsonBody["username"].getStr
        pwd = jsonBody["password"].getStr
      except: resp(Http400, makeJsonResp(Http400), JSON)

      try:
        let authToken = makeAuthToken(cfg, uname, pwd)
        resp($(%authToken), JSON)
      except: resp(Http401, makeJsonResp(Http401, getCurrentExceptionMsg()), JSON)

    get "/verify-auth":
      checkAuth(); if not authed: return true

      resp(Http200, $(%*{ "username": session.user.name }), JSON)

    get "/projects":
      ## List project summaries (ProjectDefs only)

      checkAuth(); if not authed: return true

      resp($(%cfg.projects), JSON)

    post "/projects":
      ## Create a new project definition

      checkAuth(); if not authed: return true

      # TODO
      resp(Http501, makeJsonResp(Http501), JSON)

    get "/project/@projectName":
      ## Return a project's configuration, as well as it's versions.

      checkAuth(); if not authed: return true

      # Make sure we know about that project
      var projDef: ProjectDef
      try: projDef = cfg.getProject(@"projectName")
      except: resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)

      var projConf: ProjectConfig
      try: projConf = getProjectConfig(cfg, @"projectName", "")
      except: discard ""

      let respJson = newJObject()
      respJson["definition"] = %projDef
      respJson["versions"] = %listVersions(cfg, @"projectName")
      if not projConf.name.isNil:
        respJson["latestConfig"] = %projConf

      resp(pretty(respJson), JSON)

    get "/project/@projectName/versions":
      ## Get a list of all versions that we have built

      checkAuth(); if not authed: return true

      try: resp($(%listVersions(cfg, @"projectName")), JSON)
      except:
        if getCurrentException() is KeyError:
          resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)
        else:
          when not defined(release): debug getCurrentException().getStackTrace()
          error "unable to list versions for project " & @"projectName" &
            ":\n" & getCurrentExceptionMsg()
          resp(Http500, makeJsonResp(Http500, "internal server error"), JSON)

    get "/project/@projectName/version/@version?":
      ## Get a detailed project record including step definitions (ProjectConfig).

      checkAuth(); if not authed: return true

      # Make sure we know about that project
      try: resp($(%getProjectConfig(cfg, @"projectName", @"version")), JSON)
      except: resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)

    get "/project/@projectName/runs":
      ## List all runs

      checkAuth(); if not authed: return true

      try: resp($(%listRuns(cfg, @"projectName")), JSON)
      except: resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)

    get "/project/@projectName/runs/active":
      ## List all currently active runs

      checkAuth(); if not authed: return true

      try:
        let activeRuns = workers
          .filterIt(it.process.running and it.projectName == @"projectName")
          .mapIt(cfg.getRun(@"projecName", $it.runId));
        resp($(%activeRuns), JSON)
      except:
        if getCurrentException() is KeyError:
          resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)
        else:
          when not defined(release): debug getCurrentException().getStackTrace()
          error "problem loading active runs: " & getCurrentExceptionMsg()
          resp(Http500, makeJsonResp(Http500, "internal server error"), JSON)

    get "/project/@projectName/run/@runId":
      ## Details for a specific run

      checkAuth(); if not authed: return true

      # Make sure we know about that project
      try: discard cfg.getProject(@"projectName")
      except: resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)

      if not existsRun(cfg, @"projectName", @"runId"):
        resp(Http404, makeJsonResp(Http404, "no such run for project"), JSON)

      try: resp($getRun(cfg, @"projectName", @"runId"), JSON)
      except: resp(Http500, makeJsonResp(Http500, getCurrentExceptionMsg()), JSON)

    get "/project/@projectName/step/@stepName/status/@buildRef":
      ## Get detailed information about the status of a step (assuming it has been built)

      checkAuth(); if not authed: return true

      try: resp($cfg.getBuildStatus(@"projectName", @"stepName", @"buildRef"), JSON)
      except: resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)

    post "/project/@projectName/step/@stepName/run/@buildRef?":
      # Kick off a run

      checkAuth(); if not authed: return true

      let runRequest = RunRequest(
        runId: genUUID(),
        projectName: @"projectName",
        stepName: @"stepName",
        buildRef: if @"buildRef" != "": @"buildRef" else: nil,
        timestamp: getLocalTime(getTime()),
        forceRebuild: false) # TODO support this with optional query params

      # TODO: instead of immediately spawning a worker, add the request to a
      # queue to be picked up by a worker. Allows capping the number of worker
      # prcesses, distributing, etc.
      try:
        let (status, worker) = spawnWorker(cfg, runRequest)
        workers.add(worker)

        resp($Run(
          id: runRequest.runId,
          request: runRequest,
          status: status), JSON)
      except: resp(Http404, makeJsonResp(Http404, getCurrentExceptionMsg()), JSON)

    get "/service/debug/ping":
      if not cfg.debug: resp(Http404, makeJsonResp(Http404), JSON)
      else: resp($(%"pong"), JSON)

    post "/service/debug/stop":
      if not cfg.debug: resp(Http404, makeJsonResp(Http404), JSON)
      else:
        let shutdownFut = sleepAsync(100)
        shutdownFut.callback = proc(): void = complete(stopFuture)
        resp($(%"shutting down"), JSON)

    #[
    get re".*":
      resp(Http404, makeJsonResp(Http404), JSON)

    post re".*":
      resp(Http404, makeJsonResp(Http404), JSON)
    ]#

  proc performMaintenance(cfg: StrawBossConfig): void =
    # Prune workers
    workers = workers.filterIt(it.process.running())
    debug "Performing maintanance: " & $len(workers) & " active workers after pruning."

    let fut = sleepAsync(cfg.maintenancePeriod)
    fut.callback =
      proc(): void =
        callSoon(proc(): void = performMaintenance(cfg))


  info "StrawBoss is bossing people around."

  callSoon(proc(): void = performMaintenance(cfg))
  waitFor(stopFuture)
