import cliutils, httpclient, json, os, osproc, sequtils, strutils, tempfile,
  times, unittest, untar, uuids

from langutils import sameContents

import ../testutil
import ../../../main/nim/strawbosspkg/configuration
import ../../../main/nim/strawbosspkg/core

let apiBase = "http://localhost:8180/api"
let cfgFilePath = "src/test/json/strawboss.config.json"
let cfg = loadStrawBossConfig(cfgFilePath)
let TIMEOUT = 2.minutes

# Util template intended for use to manually review test case.
# Inserting into a test case will prevent the test case from cleaning up it's
# working files and echo the command to start StrawBoss using that test's
# configuration and working files.
template keepEnv(): untyped =
  preserveEnv = true
  echo "artifacts dir: " & tempBuildDataDir
  echo "strawboss serve -c " & tempCfgPath

suite "strawboss server":

  # Suite setup: extract test project
  let testProjTempDir = mkdtemp()
  let testProjTarFile = newTarFile("src/test/test-project.tar.gz")
  let testProjName = "test-project"
  testProjTarFile.extract(testProjTempDir)

  # per-test setup: spin up a fresh strawboss instance
  setup:
    let tempBuildDataDir = mkdtemp()
    let (_, tempCfgPath) = mkstemp()
    var preserveEnv = false

    # copy our test config
    var newCfg = cfg
    newCfg.buildDataDir = tempBuildDataDir

    # update the repo string for the extracted test project
    var testProjDef = newCfg.getProject(testProjName)
    testProjDef.repo = testProjTempDir
    newCfg.setProject(testProjName, testProjDef)

    # save the updated config and start the strawboss instance using it
    writeFile(tempCfgPath, $newCfg)
    let serverProcess = startProcess("./strawboss", ".",
      @["serve", "-c", tempCfgPath], loadEnv(), {poUsePath})

    # give the server time to spin up
    sleep(100)

  teardown:
    discard newAsyncHttpClient().post(apiBase & "/service/debug/stop")

    if not preserveEnv:
      removeDir(tempBuildDataDir)
      removeFile(tempCfgPath)

    # give the server time to spin down but kill it after that
    sleep(100)
    if serverProcess.running: kill(serverProcess)

  test "handle missing project configuration":
    let http = newAuthenticatedHttpClient(apibase, "bob@builder.com", "password")
    let resp = http.get(apiBase & "/projects/" & cfg.projects[0].name)
    check resp.status.startsWith("404")

  test "gives 404 when no versions built":
    let http = newAuthenticatedHttpClient(apibase, "bob@builder.com", "password")
    let resp = http.get(apiBase & "/projects/" & testProjName & "/versions")
    check resp.status.startsWith("404")

  test "GET /api/project/@projectName/versions":
    let cachedConfsDir = tempBuildDataDir & "/" & testProjName & "/configurations"
    let expectedVersions = @["alpha", "beta", "1.0.0", "1.0.1"]

    # Touch configuration files
    createDir(cachedConfsDir)
    for v in expectedVersions:
      var f: File
      check open(f, cachedConfsDir & "/" & v & ".json", fmWrite)
      close(f)

    let http = newAuthenticatedHttpClient(apibase, "bob@builder.com", "password")
    let resp = http.get(apiBase & "/project/" & testProjName & "/versions")
    let returnedVersions = parseJson(resp.body).getElems.mapIt(it.getStr)
    check sameContents(expectedVersions, returnedVersions)

  test "run a successful build with artifacts":
    let http = newAuthenticatedHttpClient(apibase, "bob@builder.com", "password")
    let resp = http.post(apiBase & "/project/" & testProjName & "/step/build/run/0.2.1")
    check resp.status.startsWith("200")

    # Check that the run was queued
    let queuedRun = parseRun(parseJson(resp.body))
    check queuedRun.status.state == BuildState.queued

    # Wait for the build to complete
    let completedRun = http.waitForBuild(apiBase, testProjname, $queuedRun.id)

    # check that the run directory, run request, status, and output logs exist
    let runsDir = tempBuildDataDir & "/" & testProjName & "/runs"
    let runId = $completedRun.id
    check existsDir(runsDir)
    for suffix in [".request.json", ".status.json", ".stdout.log", ".stderr.log"]:
      check existsFile(runsDir & "/" & runId & suffix)

    # check that the project directory has been created in the artifacts repo
    let runArtifactsDir = tempBuildDataDir & "/" & testProjName & "/artifacts/build/0.2.1"
    check existsDir(runArtifactsDir)

    # check that the build step status file has been created
    let statusFile = tempBuildDataDir & "/" & testProjName & "/status/build/0.2.1.json"
    check fileExists(statusFile)

    # check that the status is complete
    var status = loadBuildStatus(statusFile)
    check status.state == BuildState.complete

    # check that the artifacts we expect are present
    let binFile = runArtifactsDir & "/test_project"
    check existsFile(binFile)

  test "run a multi-step build":
    let http = newAuthenticatedHttpClient(apibase, "bob@builder.com", "password")

    # Run the "test" step (depends on "build")
    var resp = http.post(apiBase & "/project/" & testProjname & "/step/test/run/0.2.1")
    check resp.status.startsWith("200")

    let queuedRun = parseRun(parseJson(resp.body))
    let completedRun = http.waitForBuild(apiBase, testProjName, $queuedRun.id)

    # there should be successful status files for both the build and test steps
    for stepName in ["build", "test"]:
      let statusFile = tempBuildDataDir & "/" & testProjName & "/status/" & stepName & "/0.2.1.json"
      check fileExists(statusFile)

      let status = loadBuildStatus(statusFile)
      check status.state == BuildState.complete

  #test "already completed steps should not be rebuilt":
  #  let http = newAuthenticatedHttpClient(apibase, "bob@builder.com", "password")
  #  let runArtifactsDir = tempBuildDataDir & "/" & testProjName & "/artifacts/build/0.2.1"
  #  let exeModTime = getLastModificationTime(runArtifactsDir & "/test_project")

    # Run the "build" step
    # Kick off a build that depends on "build" (which was run in the last test)

  # TODO
  #test "kick off multiple runs and check the list of active runs via the API":
  #  check false

  # Last-chance catch to kill the server in case some test err'ed and didn't
  # reach it's teardown handler
  discard newAsyncHttpClient().post(apiBase & "/service/debug/stop")

  # Also, delete the extracted test project "source" repo
  removeDir(testProjTempDir)

